package tech.nevets.modbot;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import tech.nevets.modbot.api.BotModule;
import tech.nevets.modbot.api.BotModuleFactory;
import tech.nevets.modbot.api.IBotModule;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.JarInputStream;
import java.util.jar.Manifest;

public class ModuleLoader {
    private static final Logger LOGGER = LoggerFactory.getLogger(ModuleLoader.class);

    public static List<Path> getModules() {
        File modulesPath = new File("./modules/");
        if (!modulesPath.exists()) {
            modulesPath.mkdirs();
        }

        List<Path> modules = new ArrayList<>();
        try {
            Files.walk(Paths.get("./modules/"))
                    .filter(Files::isRegularFile)
                    .forEach(modules::add);
        } catch (IOException e) {
            LOGGER.error("Unable to get module list");
            e.printStackTrace();
        }

        return modules;
    }

    //TODO: make this dynamically wrap all interface modules into abstract modules
    //TODO: handle both types of modules

    public static List<BotModule> loadModules() {
        List<Path> modulePaths = getModules();
        List<BotModule> modules = new ArrayList<>();

        for (Path module : modulePaths) {
            JarInputStream jarStream;
            Manifest mf;
            Attributes attributes;
            String mainClassName;
            String moduleName;

            try {
                jarStream = new JarInputStream(new FileInputStream(String.valueOf(module)));
                mf = jarStream.getManifest();
                attributes = mf.getMainAttributes();
                mainClassName = attributes.getValue("Main-Class");
                moduleName = attributes.getValue("Module-Name");
            } catch (final Exception e) {
                LOGGER.error("Error while reading module attributes from module: " + module.getFileName());
                e.printStackTrace();
                continue;
            }

            if (mainClassName == null) {
                LOGGER.error("Unable to get mainClassName from module: " + moduleName);
                continue;
            }

            File file = module.toFile();
            URL[] url;
            try {
                url = new URL[]{file.toURI().toURL()};
            } catch (MalformedURLException e) {
                LOGGER.error("Error occurred while getting file URL from module " + moduleName);
                e.printStackTrace();
                continue;
            }
            URLClassLoader urlClassLoader = URLClassLoader.newInstance(url);

            Class<?> mainClass;
            try {
                mainClass = Class.forName(mainClassName, true, urlClassLoader);
            } catch (ClassNotFoundException e) {
                LOGGER.error("Error getting class from classLoader, Class Not Found in module " + moduleName + "!");
                e.printStackTrace();
                continue;
            }

            BotModule newModule;
            String moduleType = attributes.getValue("Module-Type");
            if (moduleType.equalsIgnoreCase("inherit") || moduleType.equalsIgnoreCase("i")) {
                IBotModule iNewModule;
                try {
                    iNewModule = (IBotModule) mainClass.getDeclaredConstructor().newInstance();
                } catch (InvocationTargetException | InstantiationException | IllegalAccessException | NoSuchMethodException e) {
                    LOGGER.error("Error while creating instance of module " + moduleName);
                    e.printStackTrace();
                    continue;
                }
                newModule = BotModuleFactory.fromIBotModule(iNewModule, moduleName);
            } else if (moduleType.equalsIgnoreCase("extend") || moduleType.equalsIgnoreCase("e")) {
                try {
                    newModule = (BotModule) mainClass.getDeclaredConstructor(String.class).newInstance(moduleName);
                } catch (InvocationTargetException | InstantiationException | IllegalAccessException | NoSuchMethodException e) {
                    LOGGER.error("Error while creating instance of module " + moduleName);
                    e.printStackTrace();
                    continue;
                }
            } else {
                LOGGER.error("Error loading module: Module " + moduleName + " has an invalid Module-Type!");
                continue;
            }

            newModule.onPreEnable();
            modules.add(newModule);

            try {
                jarStream.close();
                if (urlClassLoader != null) {
                    urlClassLoader.close();
                }
            } catch (IOException e) {
                LOGGER.error("Error closing jarStream or classLoader");
                e.printStackTrace();
            }
        }

        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < modules.size(); i++) {
            sb.append(modules.get(i).getModuleName());
            if (i < modules.size()) {
                sb.append(", ");
            }
        }

        LOGGER.info("Loaded modules: " + sb);
        return modules;
    }
}
